# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionState.Login do
  require Amethyst.ConnectionState.Macros
  alias Amethyst.ConnectionState.Macros

  require Logger

  @moduledoc """
  This module contains the packets and logic for the Login state.
  """
  Macros.defpacket_clientbound :disconnect, 0x00, 767, [reason: :nbt]
  Macros.defpacket_clientbound :encryption_request, 0x01, 767, [
    server_id: :string,
    public_key: :byte_array,
    verify_token: :byte_array,
    should_authenticate: :bool
  ]
  Macros.defpacket_clientbound :login_success, 0x02, 767, [
    uuid: :uuid,
    username: :string,
    properties: {:array, [
      name: :string,
      value: :string,
      signature: {:optional, :string}
    ]},
    strict_error_handling: :bool
  ]
  Macros.defpacket_clientbound :set_compression, 0x03, 767, [threshold: :varint]
  Macros.defpacket_clientbound :login_plugin_request, 0x04, 767, [
    message_id: :varint,
    channel: :string,
    data: :raw
  ]
  Macros.defpacket_clientbound :cookie_request, 0x05, 767, [identifier: :string]

  Macros.defpacket_serverbound :login_start, 0x00, 767, [name: :string, player_uuid: :uuid]
  Macros.defpacket_serverbound :encryption_response, 0x01, 767, [
    shared_secret: :byte_array,
    verify_token: :byte_array
  ]
  Macros.defpacket_serverbound :login_plugin_response, 0x02, 767, [
    message_id: :varint,
    data: {:optional, :raw}
  ]
  Macros.defpacket_serverbound :login_acknowledged, 0x03, 767, []
  Macros.defpacket_serverbound :cookie_response, 0x04, 767, [identifier: :string, payload: {:optional, :byte_array}]

  def handle(%{packet_type: :login_start, name: name, player_uuid: player_uuid}, 767, state) do
    Logger.debug("Received login start for #{name} with UUID #{player_uuid}")
    if Application.fetch_env!(:amethyst, :encryption) do
      raise RuntimeError, "Encryption is not currently supported"
    else
      send(self(), {:send_packet, %{
        packet_type: :login_success,
        uuid: player_uuid,
        username: name,
        properties: [],
        strict_error_handling: false
      }})
      :ok
    end
  end

  def handle(%{packet_type: :login_acknowledged}, 767, _state) do
    Logger.debug("Received login acknowledged")
    send(self(), {:set_state, Amethyst.ConnectionState.Configuration})
    :ok
  end

  def disconnect(reason) do
    %{packet_type: :disconnect, reason:
      {:compound, %{
        "text" => {:string, reason}
      }}
    }
  end
end
